/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.conqat.lib.commons.collections.UnmodifiableList;
import org.eclipse.emf.common.command.CommandStack;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.interfaces.IStorageProvider;
import org.fortiss.tooling.kernel.interfaces.ITopLevelElementChangeListener;
import org.fortiss.tooling.kernel.internal.storage.eclipse.EclipseResourceStorageProvider;
import org.fortiss.tooling.kernel.services.IPersistencyService;

/**
 * This class implements the {@link IPersistencyService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @levd.rating RED Rev:
 */
public class PersistencyService implements IPersistencyService {

	/** Stores the storage providers and their element caches. */
	private final Map<IStorageProvider, List<EObject>> storageProviderCache = new HashMap<IStorageProvider, List<EObject>>();

	/** Stores the top-level element listeners. */
	private final List<ITopLevelElementChangeListener> listeners = new ArrayList<ITopLevelElementChangeListener>();

	/** Constructor. */
	public PersistencyService() {
		// TODO (FH): replace with extension mechanism
		IStorageProvider provider = new EclipseResourceStorageProvider();
		storageProviderCache.put(provider, provider.getTopLevelElements());

		refreshAllProviderElements();
	}

	/** {@inheritDoc} */
	@Override
	public UnmodifiableList<EObject> getTopLevelElements() {
		List<EObject> result = new ArrayList<EObject>();
		for (IStorageProvider provider : storageProviderCache.keySet()) {
			result.addAll(storageProviderCache.get(provider));
		}
		return CollectionUtils.asUnmodifiable(result);
	}

	/** {@inheritDoc} */
	@Override
	public void refreshTopLevelElements(IStorageProvider provider) {
		if (!storageProviderCache.containsKey(provider)) {
			return;
		}
		List<EObject> elementCache = storageProviderCache.get(provider);
		List<EObject> providedElements = provider.getTopLevelElements();
		for (EObject top : elementCache) {
			if (!providedElements.contains(top)) {
				elementCache.remove(top);
				notifyListenersAboutRemove(top);
			}
		}
		for (EObject top : providedElements) {
			if (!elementCache.contains(top)) {
				elementCache.add(top);
				notifyListenersAboutAdd(top);
			}
		}
	}

	/** Refreshes all providers' top-level elements. */
	private void refreshAllProviderElements() {
		for (IStorageProvider provider : storageProviderCache.keySet()) {
			refreshTopLevelElements(provider);
		}
	}

	/** Notifies listener about top-level element adding. */
	private void notifyListenersAboutAdd(EObject top) {
		for (ITopLevelElementChangeListener listener : listeners) {
			listener.topLevelElementAdded(top);
		}
	}

	/** Notifies listener about top-level element removal. */
	private void notifyListenersAboutRemove(EObject top) {
		for (ITopLevelElementChangeListener listener : listeners) {
			listener.topLevelElementRemoved(top);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void addTopLevelElementListener(
			ITopLevelElementChangeListener listener) {
		if (!listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeTopLevelElementListener(
			ITopLevelElementChangeListener listener) {
		listeners.remove(listener);
	}

	/** {@inheritDoc} */
	@Override
	public EObject getTopLevelElementFor(EObject modelElement) {
		while (modelElement != null) {
			for (List<EObject> topLevelElements : storageProviderCache.values()) {
				if (topLevelElements.contains(modelElement)) {
					return modelElement;
				}
			}
			modelElement = modelElement.eContainer();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IStorageProvider getStorageProviderFor(EObject modelElement) {
		while (modelElement != null) {
			for (IStorageProvider provider : storageProviderCache.keySet()) {
				if (storageProviderCache.get(provider).contains(modelElement)) {
					return provider;
				}
			}
			modelElement = modelElement.eContainer();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public CommandStack getCommandStack(EObject modelElement) {
		return getStorageProviderFor(modelElement)
				.getCommandStack(modelElement);
	}
}
